const unzip = require('node-unzip-2');
const path = require('path');

(() => {
    'use strict';

    let action = class extends odaAction {
        static get properties() {
            return {
                isActive: true,
                allowAccess: 'C'
            }
        }

        execute(param) {
            return this.init(param)
                .then(() => this.saveWorkSheetsAndWorkBookToObject())
                .then(() => this.waitingSaveAllFiles())
                .then(() => this.getSheetInfo())
                .then(() => this.transformWorkSheetsXML())
                .then(() => this.generateSourceXML())
                .then((readStream) => this.saveSourceXMLToField(readStream))
                .then(() => this.removeTemporaryFiles())
        }

        init(param) {
            this.getWorkSheets = [];
            return ODANT.findItem(param.source)
                .then((file) => {
                    this.targetFile = file;
                    return this.$item.$base.findItem(param.target)
                        .then((target) => {
                            this.targetObject = target;

                            return this.targetObject.fields
                                .then(fields => this.sourceField = fields.find(f => f.name === 'source'));
                        })
                });
        }

        saveWorkSheetsAndWorkBookToObject() {
            return new Promise((resolve) => {
                return this.targetFile.load({
                    'stream': {
                        'gzip': false
                    }
                })
                    .then(stream => {
                        stream.pipe(unzip.Parse())
                            .on('entry', (entry) => {
                                if (entry.path.startsWith('xl/worksheets/')) {
                                    this.getWorkSheets.push(this.targetObject.saveFile(entry, path.basename(entry.path)))
                                } else if (entry.path === 'xl/workbook.xml') {
                                    this.getWorkBook = this.targetObject.saveFile(entry, 'workbook.xml')
                                } else if (entry.path === 'xl/_rels/workbook.xml.rels') {
                                    this.getWorkBookXMLRels = this.targetObject.saveFile(entry, 'workbook.xml.rels.xml')
                                } else if (entry.path === 'xl/sharedStrings.xml') {
                                    this.getSharedStrings = this.targetObject.saveFile(entry, 'sharedStrings.xml')
                                } else {
                                    entry.autodrain();
                                }
                            })
                            .on('close', resolve)
                            .on('error', (err) => {
                                console.log(err)
                                resolve();
                            })

                    })
            });
        }

        waitingSaveAllFiles() {
            const saveFiles = [
                Promise.all(this.getWorkSheets),
                this.getWorkBook,
                this.getWorkBookXMLRels
            ];

            if (this.getSharedStrings) {
                saveFiles.push(this.getSharedStrings)
            }

            return Promise.all(saveFiles)
                .then(result => {
                    this.workSheets = result[0];
                    this.workBook = result[1];
                    this.workBookXMLRels = result[2];

                    if (result[3]) {
                        this.sharedStrings = result[3]
                    }
                })
        }

        generateSourceXML() {
            let wsElements = this.workSheetsTransform.reduce((a, c) => {
                a.push(`element w { attribute u {'${c.fullId.replace('~/', 'F:')}'}}`);
                return a;
            }, []);

            let xq =
                `
                    let $sequence := (
                        ${wsElements.join(',')}
                    )
                    
                   return element ROOT{
                      for $r in $sequence
                        return doc(oda:command(concat('get_file?id=/', $r/@u)))
                    } 
                `;

            return this.targetObject.API
                .POST('xquery', {
                    format: 'XML', stream:
                        {
                            'gzip': true
                        }
                }, xq)
        }

        getSheetInfo() {
            const xqGetSheetsInfo =
                `
                    declare namespace r = "http://schemas.openxmlformats.org/officeDocument/2006/relationships";
                
                    let $ws := doc(oda:command('get_file?id=${this.workBook.fullId.replace('~/', 'F:')}'))
                    let $wsRels := doc(oda:command('get_file?id=${this.workBookXMLRels.fullId.replace('~/', 'F:')}'))

                             
                        let $realRels := (
                            element x {
                                for $a in $wsRels/Relationships/Relationship
                                return element R {
                                    attribute r:id { $a/@id },
                                    $a/@*[name() != 'id']
                                }
                            }
                        )
                        
                        return 
                            for $sheet in ($ws/workbook/sheets/sheet,$realRels/R)
                            group by $sheet/@r:id
                            return element sheets{
                                attribute name { string($sheet/@name) },
                                attribute target { string($sheet/@Target) }
                            }
                `;

            return this.targetObject.API.POST('xquery', {format: 'JSON'}, xqGetSheetsInfo)
                .then(res => this.sheetInfo = res.$sheets)
        }

        transformWorkSheetsXML() {
            const workSheetsTransformPromises = this.workSheets
                .map(ws => {
                    const sheetName = path.basename(ws.Name, '.xml');
                    const sheetInfo = this.sheetInfo.find(s => s.target.includes(sheetName));
                    let xq = (this.sharedStrings)
                        ? `let $ws := doc(oda:command(\'get_file?id=${this.sharedStrings.fullId.replace('~/', 'F:')}\'))`
                        : `let $ws := ''`;

                    xq += `
                            let $data := element DATA {
                                for $a at $i in //row
                                  return element R{
                                      for $b in $a/c
                                        return 
                                            attribute {string-join(tokenize(string($b/@r[1]), '\\d'))}{
                                                if ($b/@t = 's') 
                                                then 
                                                    $ws/sst/si[oda:eval($b/v)]/t
                                                else 
                                                    ($b/v)  
                                            },
                                            attribute rowId {$i} 
                                }
                            }
                                
                            let $meta := element METADATA {
                                for $a in distinct-values($data//@*/name(.))
                                return element FIELD{
                                    attribute {'Name'} {$a}
                                }
                            }
                    
                            return element DATASET {
                                attribute name {'${sheetInfo.name}'},
                                $meta,
                                $data
                            }
                        `
                    ;

                    return ws.API.POST('xquery', {
                        format: 'XML', 'stream': {
                            'gzip': false
                        }
                    }, xq)
                        .then(stream => this.targetObject.saveFile(stream, `tr_${ws.Name}.xml`))
                });

            return Promise.all(workSheetsTransformPromises)
                .then((res) => this.workSheetsTransform = res)
                .catch((err) => console.log(err))
        }


        saveSourceXMLToField(readStream) {
            return this.sourceField.setValue(this.targetObject, new File({
                'name': 'source.xml',
                'stream': readStream
            }))
                .then((file) => {
                        this.resultXMLFile = file;

                        return this.targetObject.save()
                    }
                )
                .catch(err => console.log(err));
        }

        removeTemporaryFiles() {
            return Promise.all(this.workSheets.map(ws => ws.remove()))
                .then(() => Promise.all(this.workSheetsTransform.map(wst => wst.remove())))
                .then(() => this.resultXMLFile.fullId);
        }
    };

    action.register();
    //For testing
    if (typeof module !== 'undefined' && typeof module.exports !== 'undefined') {
        module.exports = action;
    }
})();