**listen** — это метод, который позволяет назначить обработчик одному или нескольким событиям.

Он объявляется внутри компонента следующим образом:

```javascript
listen(event='', callback, props = {target: this, once: false, useCapture: false})
```

Ему передаются 3 параметра:

1. **event** — список типов событий, для которых назначается обработчик. Если указывается сразу несколько событий, то в списке они должны разделятся запятыми. Например, **'click, contextmenu'**.
1. **callback** — обработчик, который будет выполняться при возникновении указанного события. Например, **()=>{console.log('Обработчик события')}**.
1. **props** — объект с дополнительными параметрами, управляющий регистрацией обработчика события. Его можно не указывать, тогда в этом случае будут использованы параметры по умолчанию **{target: this, once: false, useCapture: false}**.
    1. **target** — определяет HTML-элемент, для которого назначается обработчик. Если ссылка на HTML-элемент не указана, то обработчик будет зарегистрирован для всего компонента.
    1. **once** — значение **true** указывает, что обработчик должен быть вызван не более одного раза, т.е. обработчик будет автоматически удален после первого вызова.
    1. **useCapture** — определяет, для какой фазы распространения события будет зарегистрирован обработчик. Значение **true** регистрирует обработчик для фазы погружения (перехвата), **false** — для фазы всплытия.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Нажми на меня'
    },
    onTap() {
        this.label = this.label === 'Ты нажал на меня' ? 'Нажми на меня' : 'Ты нажал на меня';
    },
    ready() {
        this.listen('tap', 'onTap', {target: this.$refs.btn});
    }
});
```

Обратите внимание, что имя регистрируемого обработчика **onTap** передается методу **listen** в виде строки, а не в виде ссылки на функцию. В этом случае метод **listen** с помощью метода **bind** создает специальную связанную функцию для обработчика, которая сохраняет указатель на контекст компонента, т.е. указатель **this** внутри обработчика всегда будет ссылаться на сам компонент.

Если методу **listen** передать не имя, а ссылку на обработчик, то этот обработчик будет вызван в контексте несвязанном с компонентом.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Кто я?'
    },
    onTap() {
        this._target.$domHost.label=this.constructor.name;
    },
    ready() {
        this.listen('tap', this.onTap, {target: this.$refs.btn});
    }
});
```

В данном примере указатель **this** будет ссылаться на объект класса **odaEventTap**, от имени которого будет вызываться обработчик события **tap**.

У этого объекта есть скрытый параметр **_target**, через который можно получить доступ к самому компоненту. Но такая возможность является недокументированной и использовать ее в реальных проектах не рекомендуется. В будущем это свойство может быть изменено или даже удалено по решению разработчиков.

С помощью  метода **bind** к обработчику можно привязать контент.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Кто я?'
    },
    onTap() {
        this.label=this.constructor.name;
    },
    ready() {
        this.listen('tap', this.onTap.bind(this), {target: this.$refs.btn});
    }
});
```

Однако в этом случае будет создана анонимная обёрточная функция, в теле которой будет вызываться метод **onTap**. В результате этого будет уже невозможно отменить регистрацию этого обработчика методом **unlisten**.

Если использовать стрелочную функцию в качестве обработчика события, то контекст внутри нее будет всегда связан с самим компонентом.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Чей я?'
    },
    ready() {
        this.listen('tap', ()=>{this.label = this.constructor.name}, {target: this.$refs.btn});
    }
});
```

Если в качестве обработчика использовать анонимную функцию, то указатель **this** будет ссылаться уже не на сам компонент, а на специальный объект события, в контексте которого будет вызываться данная функция.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Кто я?'
    },
    ready() {
        this.listen('tap',function () {this._target.$domHost.label=this.constructor.name}, {target: this.$refs.btn});
    }
});
```

К анонимной функции, как и к методу, с помощью метода **bind** можно привязать контекст компонента. Методом **unlisten**, однако, отменить такой обработчик будет уже невозможно.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Кто я?'
    },
    ready() {
        this.listen('tap', (function () {this.label=this.constructor.name}).bind(this), {target: this.$refs.btn});
    }
});
```

В данном примере указатель **this** будет ссылаться на объект класса **odaEventTap**, от имени которого вызывается анонимный обработчик события **tap**.

Обработчику можно передать объект события в списке параметров.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Нажми на меня'
    },
    onTap(e) {
        this.label = e.constructor.name;
    },
    ready() {
        this.listen('tap', 'onTap', {target: this.$refs.btn});
    }
});
```

Для того чтобы событие было обработано только один раз, необходимо использовать модификатор **once**, который передается методу третьим параметром.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Я слушаю тебя'
    },
    onTap() {
        this.label = this.label === 'Я слушаю тебя' ? 'Я уже не слышу тебя' : 'Я слушаю тебя';
    },
    ready() {
        this.listen('tap','onTap', {target: this.$refs.btn, once:true});
    }
});
```

Так как модификатор **once** передается методу со значением **true**, то после первого вызова обработчика он будет автоматически удален. В результате этого все последующие нажатия на кнопку ни к каким действиям приводить уже не будут.

Если необходимо назначить один и тот же обработчик для разных событий, то в первом параметре метода **listen** можно перечислить типы событий через запятую.

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref='btn'>{{label}}</button>
    `,
    props: {
        label: 'Нажми на меня правой или левой кнопкой мыши'
    },
    onTap(e) {
        switch(e.type) {
            case 'tap': this.label = 'Нажат левая кнопка мыши'; break;
            case 'contextmenu': this.label = 'Нажат правая кнопка мыши'; break;
        }
    },
    ready() {
        this.listen('tap, contextmenu','onTap',{target: this.$refs.btn});
    }
});
```

В данном примере один и тот же обработчик будет срабатывать при нажатии как левой кнопки мыши, так и правой.

Если в объекте не указать дополнительные параметры **target** и **useCapture**, то они будут заданы по умолчанию, т.е. для компонента будет зарегистрирован слушатель (**target**=**this**), а событие будет обрабатываться при «всплытии» (**useCapture**=**false**).

Например,

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button>Нажми на меня: {{count}}</button>
        <button>Нажми на меня: {{count}}</button>
    `,
    props: {
        count: 0
    },
    onTap() {
        this.count++;
    },
    ready() {
        this.listen('tap','onTap');
    }
});
```

В данном примере нажатие мышкой на любую кнопку будет приводить к вызову одного и того же обработчика в фазе всплытия события **tap**, который будет зарегистрирован на уровне компонента по умолчанию.

Метод **listen** позволяет компоненту перехватывать события, происходящие за его пределами. Щелчком мыши по любому месту в окне примера ниже можно как показать, так и скрыть флаг. Это происходит благодаря перехвату события **click** объекта **window**.

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <span>Щелкните в окне примера, чтобы показать или скрыть флаг</span>
        <div style="width:100px; border:solid 1px" ~show="show">
            <div style="width:100px; background:white; color:white">.</div>
            <div style="width:100px; background:blue; color:blue">.</div>
            <div style="width:100px; background:red; color:red">.</div>
        </div>
    `,
    props: {
        show: true
    },
    onClick() {
        this.show = !this.show;
    },
    ready() {
        this.listen('click', 'onClick', {target: window, useCapture: true});
    }
});
```

Немного изменим пример с флагом, чтобы продемонстрировать динамическое управление перехвата события с помощью пары методов **listen** и **unlisten**. Теперь флаг будет показываться при нажатии на кнопку, а скрываться по щелчку в любой части окна примера. Обработчик события, скрывающего флаг, будет регистрироваться в момент отображения флага и отключаться, когда флаг скрыт.

```javascript _run_line_edit_[my-component.js]_h=80_
 ODA({
    is: 'my-component',
    template: `
        <button @click="clickBtn">Показать флаг</button>
        <span>Щелкните в окне, чтобы скрыть флаг</span>
        <div style="width:100px; border:solid 1px" ~show="show">
            <div style="width:100px; background:white; color:white">.</div>
            <div style="width:100px; background:blue; color:blue">.</div>
            <div style="width:100px; background:red; color:red">.</div>
        </div>
    `,
    props: {
        show: false
    },
    clickBtn() {
        this.show = true;
        this.listen('click', 'clickWindow', {target: window, useCapture: true});
    },
    clickWindow () {
        this.show = false;
        this.unlisten('click', 'clickWindow', {target: window, useCapture: true});
    }
});
```

Динамическое управление перехватом события позволяет временно исключить компонент из цепочки распространения события.

Метод **listen** позволяет задать несколько обработчиков для одного и того же события.

```javascript _run_line_edit_[my-component.js]
 ODA({
    is: 'my-component',
    template: `
        <button ref="btn">Инкремент счетчиков</button>
        <div>Счетчик 1={{count1}}</div>
        <div>Счетчик 2={{count2}}</div>
    `,
    props: {
        count1: 0,
        count2: 0
    },
    onTap1() {
        this.count1++;
    },
    onTap2() {
        this.count2++;
    },
    ready() {
        this.listen('tap', 'onTap1', {target: this.$refs.btn});
        this.listen('tap', 'onTap2', {target: this.$refs.btn});
    }
});
```

В этом случае метод **listen** необходимо будет вызвать несколько раз.
